import json
import os
from argparse import ArgumentParser
from typing import List
from rkd.core.api.contract import TaskInterface, ExecutionContext, ExtendableTaskInterface
from rkd.core.api.inputoutput import IO
from rkd.core.api.syntax import TaskDeclaration
from rkd.core.execution.lifecycle import CompilationLifecycleEvent


class ComposerScriptTask(TaskInterface):
    group_name: str
    task_name: str
    composer_task_name: str

    def __init__(self, group_name: str, task_name: str, composer_task_name: str):
        self.group_name = group_name
        self.task_name = task_name
        self.composer_task_name = composer_task_name

    def get_name(self) -> str:
        return self.task_name

    def get_group_name(self) -> str:
        return self.group_name

    def get_description(self) -> str:
        return 'Composer task'

    def execute(self, context: ExecutionContext) -> bool:
        if context.get_arg('--clear'):
            self.sh('rm -rf vendor')

        if not os.path.isdir('vendor') or context.get_arg('--install'):
            install_args = ' --no-progress '

            if context.get_arg('--no-dev'):
                install_args += ' --no-dev '

            if context.get_arg('--no-scripts'):
                install_args += ' --no-scripts '

            self.sh(f'composer install {install_args}')

        self.sh('composer run {task_name} -- {args}'.format(
            task_name=self.composer_task_name,
            args=''
            # args=context.get_unknown_args()
        ))

        return True

    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument('--no-dev', help='Disables installation of require-dev packages', action='store_true')
        parser.add_argument('--no-scripts', help='Skips the execution of all scripts defined in composer.json file',
                            action='store_true')
        parser.add_argument('--install', help='Enforce `composer install`', action='store_true')
        parser.add_argument('--clear', help='Force remove `vendor` directory first', action='store_true')


class ComposerIntegrationTask(ExtendableTaskInterface):
    """Runs tasks from composer.json"""

    def get_name(self) -> str:
        return ':composer'

    def get_group_name(self) -> str:
        return ':php'

    def configure_argparse(self, parser: ArgumentParser):
        pass

    @staticmethod
    def find_composer_tasks(io: IO) -> List[str]:
        if not os.path.isfile('composer.json'):
            io.debug('composer.json not found')
            return []

        try:
            io.debug('Trying to load composer.json')

            with open('composer.json', 'r') as f:
                data = json.loads(f.read())

            return data['scripts'].keys()

        except Exception as exc:
            # todo specific exception class
            raise Exception('Cannot load composer.json') from exc

    def compile(self, event: CompilationLifecycleEvent) -> None:
        """
        Collects all scripts from "composer.json" and adds into the RKD's context as tasks
        :param event:
        :return:
        """

        tasks: List[TaskDeclaration] = []

        for task_name in self.find_composer_tasks(event.io):
            tasks.append(
                TaskDeclaration(ComposerScriptTask(
                    group_name='',
                    task_name=event.get_current_declaration().to_full_name() + ':' + task_name,
                    composer_task_name=task_name
                ))
            )

        event.expand_into_group(tasks, pipeline=False)

    def execute(self, context: ExecutionContext) -> bool:
        for task_name in self.find_composer_tasks(self.io()):
            self.io().outln(task_name)

        return True




import os
import subprocess
import tempfile
from argparse import ArgumentParser
from copy import copy
from typing import Dict, Union, Optional, List

from rkd.core.api.inputoutput import ReadableStreamType
from rkd.core.api.contract import ExecutionContext, ArgumentEnv, MultiStepLanguageExtensionInterface
from rkd.core.api.syntax import TaskDeclaration
from rkd.core.execution.lifecycle import ConfigurationLifecycleEvent
from rkd.core.standardlib.docker import RunInContainerBaseTask


class PhpScriptTask(RunInContainerBaseTask):
    """
    # <sphinx:extending-tasks>
    Execute a PHP code (using a docker container)
    Can be extended - this is a base task.

    Inherits settings from `RunInContainerBaseTask`.

    **Configuration:**

        - script: Path to script to load instead of stdin (could be a relative path)
        - version: PHP version. Leave None to use default 8.0-alpine version

    **Example of usage:**

        .. code:: yaml

            version: org.riotkit.rkd/yaml/v2
            imports:
                - rkd.php.script.PhpScriptTask
            tasks:
                :yaml:test:php:
                    extends: rkd.php.script.PhpScriptTask
                    configure@before_parent: |
                        self.version = '7.2-alpine'
                    inner_execute@after_parent: |
                        self.in_container('php --version')
                        print('IM AFTER PARENT. At first the PHP code from "input" will be executed.')
                        return True
                    input: |
                        var_dump(getcwd());
                        var_dump(phpversion());

    **Example of usage with MultiStepLanguageAgnosticTask:**

        .. code:: yaml

            version: org.riotkit.rkd/yaml/v1
            tasks:
                :exec:
                    environment:
                        PHP: '7.4'
                        IMAGE: 'php'
                    steps: |
                        #!rkd.php.script.PhpLanguage
                        phpinfo();

    # </sphinx:extending-tasks>
    """

    script: Optional[str]
    version: Optional[str]
    name: Optional[str]
    input: Optional[callable]

    def __init__(self):
        super().__init__()
        self.user = 'www-data'
        self.entrypoint = 'sleep'
        self.command = '9999999'
        self.script = None
        self.version = None
        self.name = None
        self.input = None

    def get_name(self) -> str:
        return ':php' if not self.name else self.name

    def get_group_name(self) -> str:
        return ''

    @classmethod
    def get_declared_envs(cls) -> Dict[str, Union[str, ArgumentEnv]]:
        return {
            'PHP': ArgumentEnv('PHP', '--php', '8.0-alpine'),
            'IMAGE': ArgumentEnv('IMAGE', '--image', 'php')
        }

    def configure(self, event: ConfigurationLifecycleEvent) -> None:
        super().configure(event)

        self.docker_image = '{image}:{version}'.format(
            image=event.ctx.get_arg_or_env('--image'),
            version=self.version if self.version else event.ctx.get_arg_or_env('--php')
        )

        # todo: Check - is workdir already set there? (subprojects, custom workdir etc.)
        self.mount(local=os.getcwd(), remote=os.getcwd())

    def inner_execute(self, context: ExecutionContext) -> bool:
        """
        Execute a code when the container is up and running
        :param context:
        :return:
        """

        try:
            # takes a RKD task input as input file, stdin is interactive
            if not self.script and self.get_input(context):
                input_php_code = self.get_input(context).read()

                if "<?php" not in input_php_code:
                    input_php_code = "<?php\n" + input_php_code

                with tempfile.NamedTemporaryFile() as tmp_file:
                    tmp_file.write(input_php_code.encode('utf-8'))
                    tmp_file.flush()

                    self.copy_to_container(local=tmp_file.name, remote='/tmp/script.php')      # copy file
                    self.in_container('chown www-data:www-data /tmp/script.php', user='root')  # fix permissions

                self.in_container('php /tmp/script.php', workdir=os.getcwd(), user=self.user)
                return True

            # takes stdin as input
            self.in_container(f'php {self.script}')

        except subprocess.CalledProcessError:
            self.io().error('PHP process exited with non-zero exit code')
            return False

        return True

    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument('--php', help='PHP version ("php" docker image tag)', default='8.0-alpine')
        parser.add_argument('--image', help='Docker image name', default='php')

    def get_input(self, ctx: ExecutionContext):
        if self.input:
            return self.input

        return ctx.get_input()


class PhpLanguage(PhpScriptTask, MultiStepLanguageExtensionInterface):
    """
    Language extension for MultiStepLanguageAgnosticTask

    .. code:: yaml

        version: org.riotkit.rkd/yaml/v1
        tasks:
            :exec:
                environment:
                    PHP: '7.4'
                    IMAGE: 'php'
                steps: |
                    #!rkd.php.script.PhpLanguage
                    phpinfo();

    """

    def with_predefined_details(self, code: str, name: str, step_num: int) -> 'PhpScriptTask':
        clone = copy(self)
        clone.name = name
        clone.input = ReadableStreamType(code)

        return clone


def imports() -> List[TaskDeclaration]:
    return [
        TaskDeclaration(PhpScriptTask(), internal=True)
    ]



from rkd.core.api.syntax import TaskDeclaration
from .composer import ComposerIntegrationTask
from .script import PhpScriptTask, imports as script_imports


def imports():
    return [
        TaskDeclaration(ComposerIntegrationTask())
    ] + script_imports()
